# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy
from multiprocessing import Manager


class InMemoryJournalWriter:
    def __init__(self):
        # Share the list of objects across processes, for RemoteAPI tests.
        self.manager = Manager()
        self.objects = self.manager.list()

    def write_addition(self, object_type, object_):
        self.objects.append((object_type, copy.deepcopy(object_)))

    write_update = write_addition

    def write_additions(self, object_type, objects):
        for object_ in objects:
            self.write_addition(object_type, object_)


def get_journal_writer(cls, args={}):
    if cls == 'inmemory':
        JournalWriter = InMemoryJournalWriter
    elif cls == 'kafka':
        from swh.journal.direct_writer import DirectKafkaWriter \
            as JournalWriter
    else:
        raise ValueError('Unknown storage class `%s`' % cls)

    return JournalWriter(**args)



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from .server import make_app_from_configfile

application = make_app_from_configfile()



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest

from swh.storage import converters


class TestConverters(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_db_to_author(self):
        # when
        actual_author = converters.db_to_author(
            1, b'fullname', b'name', b'email')

        # then
        self.assertEqual(actual_author, {
            'id': 1,
            'fullname': b'fullname',
            'name': b'name',
            'email': b'email',
        })

    def test_db_to_revision(self):
        # when
        actual_revision = converters.db_to_revision({
            'id': 'revision-id',
            'date': None,
            'date_offset': None,
            'date_neg_utc_offset': None,
            'committer_date': None,
            'committer_date_offset': None,
            'committer_date_neg_utc_offset': None,
            'type': 'rev',
            'directory': b'dir-sha1',
            'message': b'commit message',
            'author_id': 'auth-id',
            'author_fullname': b'auth-fullname',
            'author_name': b'auth-name',
            'author_email': b'auth-email',
            'committer_id': 'comm-id',
            'committer_fullname': b'comm-fullname',
            'committer_name': b'comm-name',
            'committer_email': b'comm-email',
            'metadata': {},
            'synthetic': False,
            'parents': [123, 456]
        })

        # then
        self.assertEqual(actual_revision, {
            'id': 'revision-id',
            'author': {
                'id': 'auth-id',
                'fullname': b'auth-fullname',
                'name': b'auth-name',
                'email': b'auth-email',
            },
            'date': None,
            'committer': {
                'id': 'comm-id',
                'fullname': b'comm-fullname',
                'name': b'comm-name',
                'email': b'comm-email',
            },
            'committer_date': None,
            'type': 'rev',
            'directory': b'dir-sha1',
            'message': b'commit message',
            'metadata': {},
            'synthetic': False,
            'parents': [123, 456],
        })

    def test_db_to_release(self):
        # when
        actual_release = converters.db_to_release({
            'id': b'release-id',
            'target': b'revision-id',
            'target_type': 'revision',
            'date': None,
            'date_offset': None,
            'date_neg_utc_offset': None,
            'name': b'release-name',
            'comment': b'release comment',
            'synthetic': True,
            'author_id': 'auth-id',
            'author_fullname': b'auth-fullname',
            'author_name': b'auth-name',
            'author_email': b'auth-email',
        })

        # then
        self.assertEqual(actual_release, {
            'author': {
                'id': 'auth-id',
                'fullname': b'auth-fullname',
                'name': b'auth-name',
                'email': b'auth-email',
            },
            'date': None,
            'id': b'release-id',
            'name': b'release-name',
            'message': b'release comment',
            'synthetic': True,
            'target': b'revision-id',
            'target_type': 'revision'
        })

    def test_db_to_git_headers(self):
        raw_data = [
            ['gpgsig', b'garbage\x89a\x43b\x14'],
            ['extra', [b'fo\\\\\\o', b'bar\\', b'inval\\\\\x99id']],
        ]

        db_data = converters.git_headers_to_db(raw_data)
        loop = converters.db_to_git_headers(db_data)
        self.assertEqual(raw_data, loop)



# Copyright (C) 2015-2017  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import unittest

import pytest

from swh.core.db.tests.db_testing import SingleDbTestFixture
from swh.model.hashutil import hash_to_bytes
from swh.storage.db import Db
from . import SQL_DIR


@pytest.mark.db
class TestDb(SingleDbTestFixture, unittest.TestCase):
    TEST_DB_NAME = 'softwareheritage-test-storage'
    TEST_DB_DUMP = os.path.join(SQL_DIR, '*.sql')

    def setUp(self):
        super().setUp()
        self.db = Db(self.conn)

    def tearDown(self):
        self.db.conn.close()
        super().tearDown()

    def test_add_content(self):
        cur = self.cursor
        sha1 = hash_to_bytes('34973274ccef6ab4dfaaf86599792fa9c3fe4689')
        self.db.mktemp('content', cur)
        self.db.copy_to([{
            'sha1': sha1,
            'sha1_git': hash_to_bytes(
                'd81cc0710eb6cf9efd5b920a8453e1e07157b6cd'),
            'sha256': hash_to_bytes(
                '673650f936cb3b0a2f93ce09d81be107'
                '48b1b203c19e8176b4eefc1964a0cf3a'),
            'blake2s256': hash_to_bytes('69217a3079908094e11121d042354a7c'
                                        '1f55b6482ca1a51e1b250dfd1ed0eef9'),
            'length': 3}],
                        'tmp_content',
                        ['sha1', 'sha1_git', 'sha256', 'blake2s256', 'length'],
                        cur)
        self.db.content_add_from_temp(cur)
        self.cursor.execute('SELECT sha1 FROM content WHERE sha1 = %s',
                            (sha1,))
        self.assertEqual(self.cursor.fetchone()[0], sha1)



# Copyright (C) 2017 the Software Heritage developers
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import binascii
from collections import defaultdict
import datetime

from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    Enum,
    ForeignKey,
    Integer,
    LargeBinary,
    String,
    Table,
    UniqueConstraint,
)

try:
    from sqlalchemy import JSON
except ImportError:
    # SQLAlchemy < 1.1
    from sqlalchemy.dialects.postgresql import JSONB as JSON

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

SQLBase = declarative_base()


class Distribution(SQLBase):
    """A distribution (e.g. Debian, Ubuntu, Fedora, ...)"""
    __tablename__ = 'distribution'

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True, nullable=False)
    type = Column(Enum('deb', 'rpm', name='distribution_types'),
                  nullable=False)
    mirror_uri = Column(String, nullable=False)

    areas = relationship('Area', back_populates='distribution')

    def origin_for_package(self, package_name, package_versions):
        """Return the origin dictionary for the given package"""
        return {
            'type': self.type,
            'url': '%s://%s/packages/%s' % (
                self.type, self.name, package_name
            ),
        }

    def __repr__(self):
        return 'Distribution(%s (%s) on %s)' % (
            self.name,
            self.type,
            self.mirror_uri,
        )


class Area(SQLBase):
    __tablename__ = 'area'
    __table_args__ = (
        UniqueConstraint('distribution_id', 'name'),
    )

    id = Column(Integer, primary_key=True)
    distribution_id = Column(Integer, ForeignKey('distribution.id'),
                             nullable=False)
    name = Column(String, nullable=False)
    active = Column(Boolean, nullable=False, default=True)

    distribution = relationship('Distribution', back_populates='areas')

    def index_uris(self):
        """Get possible URIs for this component's package index"""
        if self.distribution.type == 'deb':
            compression_exts = ('xz', 'bz2', 'gz', None)
            base_uri = '%s/dists/%s/source/Sources' % (
                self.distribution.mirror_uri,
                self.name,
            )
            for ext in compression_exts:
                if ext:
                    yield (base_uri + '.' + ext, ext)
                else:
                    yield (base_uri, None)

        raise NotImplementedError(
            'Do not know how to build index URI for Distribution type %s' %
            self.distribution.type
        )

    def __repr__(self):
        return 'Area(%s of %s)' % (
            self.name,
            self.distribution.name,
        )


class Package(SQLBase):
    __tablename__ = 'package'
    __table_args__ = (
        UniqueConstraint('area_id', 'name', 'version'),
    )

    id = Column(Integer, primary_key=True)
    area_id = Column(Integer, ForeignKey('area.id'), nullable=False)
    name = Column(String, nullable=False)
    version = Column(String, nullable=False)
    directory = Column(String, nullable=False)
    files = Column(JSON, nullable=False)

    origin_id = Column(Integer)
    task_id = Column(Integer)

    revision_id = Column(LargeBinary(20))

    area = relationship('Area')

    @property
    def distribution(self):
        return self.area.distribution

    def fetch_uri(self, filename):
        """Get the URI to fetch the `filename` file associated with the
        package"""
        if self.distribution.type == 'deb':
            return '%s/%s/%s' % (
                self.distribution.mirror_uri,
                self.directory,
                filename,
            )
        else:
            raise NotImplementedError(
                'Do not know how to build fetch URI for Distribution type %s' %
                self.distribution.type
            )

    def loader_dict(self):
        ret = {
            'id': self.id,
            'name': self.name,
            'version': self.version,
        }
        if self.revision_id:
            ret['revision_id'] = binascii.hexlify(self.revision_id).decode()
        else:
            files = {
                name: checksums.copy()
                for name, checksums in self.files.items()
            }
            for name in files:
                files[name]['uri'] = self.fetch_uri(name)

            ret.update({
                'revision_id': None,
                'files': files,
            })
        return ret

    def __repr__(self):
        return 'Package(%s_%s of %s %s)' % (
            self.name,
            self.version,
            self.distribution.name,
            self.area.name,
        )


class DistributionSnapshot(SQLBase):
    __tablename__ = 'distribution_snapshot'

    id = Column(Integer, primary_key=True)
    date = Column(DateTime, nullable=False, index=True)
    distribution_id = Column(Integer,
                             ForeignKey('distribution.id'),
                             nullable=False)

    distribution = relationship('Distribution')
    areas = relationship('AreaSnapshot', back_populates='snapshot')

    def task_for_package(self, package_name, package_versions):
        """Return the task dictionary for the given list of package versions"""
        origin = self.distribution.origin_for_package(
            package_name, package_versions,
        )

        return {
            'policy': 'oneshot',
            'type': 'load-%s-package' % self.distribution.type,
            'next_run': datetime.datetime.now(tz=datetime.timezone.utc),
            'arguments': {
                'args': [],
                'kwargs': {
                    'origin': origin,
                    'date': self.date.isoformat(),
                    'packages': package_versions,
                },
            }
        }

    def get_packages(self):
        packages = defaultdict(dict)
        for area_snapshot in self.areas:
            area_name = area_snapshot.area.name
            for package in area_snapshot.packages:
                ref_name = '%s/%s' % (area_name, package.version)
                packages[package.name][ref_name] = package.loader_dict()

        return packages


area_snapshot_package_assoc = Table(
    'area_snapshot_package', SQLBase.metadata,
    Column('area_snapshot_id', Integer, ForeignKey('area_snapshot.id'),
           nullable=False),
    Column('package_id', Integer, ForeignKey('package.id'),
           nullable=False),
)


class AreaSnapshot(SQLBase):
    __tablename__ = 'area_snapshot'

    id = Column(Integer, primary_key=True)
    snapshot_id = Column(Integer,
                         ForeignKey('distribution_snapshot.id'),
                         nullable=False)
    area_id = Column(Integer,
                     ForeignKey('area.id'),
                     nullable=False)

    snapshot = relationship('DistributionSnapshot', back_populates='areas')
    area = relationship('Area')
    packages = relationship('Package', secondary=area_snapshot_package_assoc)


class TempPackage(SQLBase):
    __tablename__ = 'temp_package'
    __table_args__ = {
        'prefixes': ['TEMPORARY'],
    }

    id = Column(Integer, primary_key=True)
    area_id = Column(Integer)
    name = Column(String)
    version = Column(String)
